# losses.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
# Adapted from code by Alex Tseng

"""
This module contains the losses used by BPNet for training.
"""

import torch


def _mixture_loss(y, y_hat_logits, y_hat_logcounts, count_loss_weight, 
	labels=None):
	"""A function that takes in predictions and truth and returns the loss.
	
	This function takes in the observed integer read counts, the predicted logits,
	and the predicted logcounts, and returns the total loss. Importantly, this
	calculates a single multinomial over all strands in the tracks and a single
	count loss across all tracks.
	
	The logits do not have to be normalized.
	
	
	Parameters
	----------
	y: torch.Tensor, shape=(n,
	"""
	
	y_hat_logits = y_hat_logits.reshape(y_hat_logits.shape[0], -1)
	y_hat_logits = torch.nn.functional.log_softmax(y_hat_logits, dim=-1)

	y = y.reshape(y.shape[0], -1)
	y_ = y.sum(dim=-1).reshape(y.shape[0], 1)

	# Calculate the profile and count losses
	if labels is not None:
		profile_loss = MNLLLoss(y_hat_logits[labels == 1], y[labels == 1]).mean()
	else:
		profile_loss = MNLLLoss(y_hat_logits, y).mean()

	count_loss = log1pMSELoss(y_hat_logcounts, y_).mean()

	# Extract the profile loss for logging
	profile_loss_ = profile_loss.item()
	count_loss_ = count_loss.item()

	# Mix losses together
	loss = profile_loss + count_loss_weight * count_loss
	
	return profile_loss_, count_loss_, loss


def MNLLLoss(logps, true_counts):
	"""A loss function based on the multinomial negative log-likelihood.

	This loss function takes in a tensor of normalized log probabilities such
	that the logsumexp across the last dimension is equal to 0 (i.e., the
	sum of the exponentiated values is equal to 1), and a tensor of true
	integer counts, and returns the log probability of observing those counts
	given the predicted distributions.

	This function can accept tensors with any number of dimensions and calculates
	the loss across the last dimension. For example, if they are both A x B x L
	arrays, then the correlation of corresponding L-arrays will be computed and 
	returned in an A x B array.

	An important note about this loss is that, despite performing very well for
	these basepair resolution models, there is nothing spatial about this loss.
	Each position in the last dimension is viewed as an independent category, and
	having high counts one basepair away is just as bad as having it 100 basepairs
	away. There is nothing wrong with this, particularly if you care strongly
	about basepair resolution, but you should keep this in mind.

	Adapted from Alex Tseng.


	Parameters
	----------
	logps: torch.Tensor, shape=(n, ..., L)
		A tensor with `n` examples and `L` possible categories. 

	true_counts: torch.Tensor, shape=(n, ..., L)
		A tensor with `n` examples and `L` possible categories.


	Returns
	-------
	loss: torch.Tensor, shape=(n, ...)
		The multinomial log likelihood loss of the true counts given the
		predicted probabilities
	"""

	log_fact_sum = torch.lgamma(torch.sum(true_counts, dim=-1) + 1)
	log_prod_fact = torch.sum(torch.lgamma(true_counts + 1), dim=-1)
	log_prod_exp = torch.sum(true_counts * logps, dim=-1)
	return -log_fact_sum + log_prod_fact - log_prod_exp


def log1pMSELoss(log_predicted_counts, true_counts):
	"""A MSE loss on the log(x+1) of the inputs.

	This loss will accept tensors of predicted counts and a vector of true
	counts and return the MSE on the log of the labels. The squared error
	is calculated for each position in the tensor and then averaged, regardless
	of the shape.

	Note: The predicted counts are in log space but the true counts are in the
	original count space.


	Parameters
	----------
	log_predicted_counts: torch.tensor, shape=(n, ...)
		A tensor of log predicted counts where the first axis is the number of
		examples. Important: these values are already in log space.

	true_counts: torch.tensor, shape=(n, ...)
		A tensor of the true counts where the first axis is the number of
		examples.


	Returns
	-------
	loss: torch.tensor, shape=(n, 1)
		The MSE loss on the log of the two inputs, averaged over all examples
		and all other dimensions.
	"""

	log_true = torch.log(true_counts+1)
	return torch.mean(torch.square(log_true - log_predicted_counts), dim=-1)



# chrombpnet.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import h5py
import time
import numpy
import torch

from .bpnet import BPNet
from .losses import MNLLLoss
from .losses import log1pMSELoss
from .performance import calculate_performance_measures
from .logging import Logger

from tqdm import trange

from tangermeme.predict import predict


class _Exp(torch.nn.Module):
	def __init__(self):
		super(_Exp, self).__init__()

	def forward(self, X):
		return torch.exp(X)


class _Log(torch.nn.Module):
	def __init__(self):
		super(_Log, self).__init__()

	def forward(self, X):
		return torch.log(X)


class ChromBPNet(torch.nn.Module):
	"""A ChromBPNet model.

	ChromBPNet is an extension of BPNet to handle chromatin accessibility data,
	in contrast to the protein binding data that BPNet handles. The distinction
	between these data types is that an enzyme used in DNase-seq and ATAC-seq
	experiments itself has a soft sequence preference, meaning that the
	strength of the signal is driven by real biology but that the exact read
	mapping locations are driven by the soft sequence bias of the enzyme.

	ChromBPNet handles this by treating the data using two models: a bias
	model that is initially trained on background (non-peak) regions where
	the bias dominates, and an accessibility model that is subsequently trained
	using a frozen version of the bias model. The bias model learns to remove
	the enzyme bias so that the accessibility model can learn real motifs.


	Parameters
	----------
	bias: torch.nn.Module 
		This model takes in sequence and outputs the shape one would expect in 
		ATAC-seq data due to Tn5 bias alone. This is usually a BPNet model
		from the bpnet-lite repo that has been trained on GC-matched non-peak
		regions.

	accessibility: torch.nn.Module
		This model takes in sequence and outputs the accessibility one would 
		expect due to the components of the sequence, but also takes in a cell 
		representation which modifies the parameters of the model, hence, 
		"dynamic." This model is usually a DynamicBPNet model, defined below.

	name: str
		The name to prepend when saving the file.
	"""

	def __init__(self, bias, accessibility, name):
		super(ChromBPNet, self).__init__()
		for parameter in bias.parameters():
			parameter.requires_grad = False

		self.bias = bias
		self.accessibility = accessibility
		self.name = name
		self.logger = None
		self.n_control_tracks = accessibility.n_control_tracks
		self.n_outputs = 1
		self._log = _Log()
		self._exp1 = _Exp()
		self._exp2 = _Exp()


	def forward(self, X, X_ctl=None):
		"""A forward pass through the network.

		This function is usually accessed through calling the model, e.g.
		doing `model(x)`. The method defines how inputs are transformed into
		the outputs through interactions with each of the layers.


		Parameters
		----------
		X: torch.tensor, shape=(-1, 4, 2114)
			A one-hot encoded sequence tensor.

		X_ctl: ignore
			An ignored parameter for consistency with attribution functions.


		Returns
		-------
		y_profile: torch.tensor, shape=(-1, 1000)
			The predicted logit profile for each example. Note that this is not
			a normalized value.
		"""

		acc_profile, acc_counts = self.accessibility(X)
		bias_profile, bias_counts = self.bias(X)

		n0, n1 = acc_profile.shape[-1], bias_profile.shape[-1]
		w = (n1 - n0) // 2
		
		y_profile = acc_profile + bias_profile[:, :, w:-w]
		y_counts = self._log(self._exp1(acc_counts) + self._exp2(bias_counts))
		
		return y_profile, y_counts


	def fit(self, training_data, optimizer, X_valid=None, y_valid=None,
		max_epochs=100, batch_size=64, validation_iter=100, dtype='float32',
		device='cuda', early_stopping=None, verbose=True):
		"""Fit the ChromBPNet model to data.

		Specifically, this function will fit the accessibility model to
		observed chromatin accessibility data, and assume that the bias model
		is frozen and pre-trained. Hence, the only parameters being trained
		in this function are those in the accessibility model.

		This function will save the best full ChromBPNet model, as well as the
		best accessibility model, found during training.


		Parameter
		---------
		training_data: torch.utils.data.DataLoader
			A data set that generates one-hot encoded sequence as input and
			read count signal for the output.

		optimizer: torch.optim.Optimizer
			A PyTorch optimizer.

		X_valid: torch.Tensor or None, shape=(-1, 4, length)
			A tensor of one-hot encoded sequences to use as input for the
			validation steps. If None, do not do validation. Default is None.

		y_valid: torch.Tensor or None, shape=(-1, 1, length)
			A tensor of read counts matched with the `X_valid` input. If None,
			do not do validation. Default is None.

		max_epochs: int
			The maximum number of training epochs to perform before stopping.
			Default is 100.

		batch_size: int
			The number of examples to use in each batch. Default is 64.

		validation_iter: int
			The number of training batches to perform before doing another
			round of validation. Set higher to spend a higher percentage of
			time in the training step.

		dtype: str or torch.dtype
			The torch.dtype to use when training. Usually, this will be torch.float32
			or torch.bfloat16. Default is torch.float32.
		
		device: str
			The device to use for training and inference. Typically, this will be
			'cuda' but can be anything supported by torch. Default is 'cuda'.

		early_stopping: int or None
			Whether to stop training early. If None, continue training until
			max_epochs is reached. If an integer, continue training until that
			number of `validation_iter` ticks has been hit without improvement
			in performance. Default is None.

		verbose: bool
			Whether to print the log as it is being generated. A log will
			be returned at the end of training regardless of this option, but
			when False, nothing will be printed to the screen during training.
			Default is False
		"""

		print("Warning: BPNet and ChromBPNet models trained using bpnet-lite may underperform those trained using the official repositories. See the GitHub README for further documentation.")
		
		dtype = getattr(torch, dtype) if isinstance(dtype, str) else dtype
		
		y_bias_profile, y_bias_counts = predict(self.bias, X_valid, 
			batch_size=batch_size, dtype=dtype, device='cuda')

		self.logger = Logger(["Epoch", "Iteration", "Training Time",
			"Validation Time", "Training MNLL", "Training Count MSE", 
			"Validation MNLL", "Validation Profile Correlation", 
			"Validation Count Pearson", "Validation Count MSE", "Saved?"], 
			verbose=verbose)

		early_stop_count = 0
		start, best_loss = time.time(), float("inf")
		
		self.logger.start()
		self.bias.eval()
		for epoch in range(max_epochs):
			for iteration, (X, y) in enumerate(training_data):
				self.accessibility.train()

				X = X.cuda().float()
				y = y.cuda()

				optimizer.zero_grad()

				with torch.autocast(device_type=device, dtype=dtype):
					acc_profile, acc_counts = self.accessibility(X)
					bias_profile, bias_counts = self.bias(X)

					y_profile = torch.nn.functional.log_softmax(acc_profile +
						bias_profile, dim=-1)

					y_counts = torch.logsumexp(torch.stack([acc_counts, 
						bias_counts]), dim=0)

					profile_loss = MNLLLoss(y_profile, y).mean()
					count_loss = log1pMSELoss(y_counts, y.sum(dim=-1).reshape(-1, 
						1)).mean()

					profile_loss_ = profile_loss.item()
					count_loss_ = count_loss.item()

					loss = profile_loss + self.accessibility.alpha * count_loss
					loss.backward()
					optimizer.step()

				if verbose and iteration % validation_iter == 0:
					train_time = time.time() - start
					tic = time.time()

					with torch.no_grad():
						self.accessibility.eval()

						y_profile, y_counts = predict(self.accessibility, 
							X_valid, batch_size=batch_size, device=device,
							dtype=dtype)

						y_profile = torch.nn.functional.log_softmax(
							y_profile + y_bias_profile, dim=-1)

						y_counts = torch.logsumexp(torch.stack([y_counts,
							y_bias_counts]), dim=0)

						measures = calculate_performance_measures(y_profile, 
							y_valid, y_counts, kernel_sigma=7, 
							kernel_width=81, measures=['profile_mnll', 
							'profile_pearson', 'count_pearson', 'count_mse'])

						profile_corr = measures['profile_pearson']
						count_corr = measures['count_pearson']

						valid_loss = measures['profile_mnll'].mean()
						valid_loss += self.accessibility.alpha * measures['count_mse'].mean()
						valid_time = time.time() - tic

						self.logger.add([epoch, iteration, train_time, 
							valid_time, profile_loss_, count_loss_, 
							measures['profile_mnll'].mean().item(), 
							numpy.nan_to_num(profile_corr).mean(),
							numpy.nan_to_num(count_corr).mean(), 
							measures['count_mse'].mean().item(),
							(valid_loss < best_loss).item()])

						if valid_loss < best_loss:
							torch.save(self, "{}.torch".format(self.name))
							torch.save(self.accessibility, 
								"{}.accessibility.torch".format(self.name))
							best_loss = valid_loss
							early_stop_count = 0
						else:
							early_stop_count += 1

					start = time.time()

				if early_stopping is not None and early_stop_count >= early_stopping:
					break

			self.logger.save("{}.log".format(self.name))

			if early_stopping is not None and early_stop_count >= early_stopping:
				break

		torch.save(self, "{}.final.torch".format(self.name))
		torch.save(self, "{}.accessibility.final.torch".format(self.name))


	@classmethod
	def from_chrombpnet_lite(self, bias_model, accessibility_model, name):
		"""Load a ChromBPNet model trained in ChromBPNet-lite.

		Confusingly, ChromBPNet-lite is a package written by Surag Nair that
		reorganized the ChromBPNet library and then was reintegrated back
		into it. However, some ChromBPNet models are still around that were
		trained using this package and this is a method for loading those
		models, not the models trained using the ChromBPNet package and not
		ChromBPNet models trained using this package.

		This method takes in paths to a h5 file containing the weights of the
		bias model and the accessibility model, both trained and whose outputs
		are organized according to TensorFlow. The weights are loaded and
		shaped into a PyTorch model and can be used as such.


		Parameters
		----------
		bias model: str
			The filename of the bias model.

		accessibility_model: str
			The filename of the accessibility model.

		name: str
			The name to use when training the model and outputting to a file.
		

		Returns
		-------
		model: bpnetlite.models.ChromBPNet
			A PyTorch ChromBPNet model compatible with the bpnet-lite package.
		"""

		bias = BPNet.from_chrombpnet_lite(bias_model)
		acc = BPNet.from_chrombpnet_lite(accessibility_model)
		return ChromBPNet(bias, acc, name)


	@classmethod
	def from_chrombpnet(self, bias_model, accessibility_model, name):
		"""Load a ChromBPNet model trained using the official repository.

		This method takes in the path to a .h5 file containing the full model,
		i.e., the bias model AND the accessibility model. If you have two
		files -- one for the bias model, and one for the accessibility model --
		load them up as separate BPNet models and create a ChromBPNet object
		afterwards.


		Parameters
		----------
		bias model: str
			The filename of the bias model.

		accessibility_model: str
			The filename of the accessibility model.

		name: str
			The name to use when training the model and outputting to a file.
		

		Returns
		-------
		model: bpnetlite.models.ChromBPNet
			A PyTorch ChromBPNet model compatible with the bpnet-lite package.
		"""


		bias = BPNet.from_chrombpnet(bias_model)
		acc = BPNet.from_chrombpnet(accessibility_model)
		return ChromBPNet(bias, acc, name)


# attribute.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>


from bpnetlite.bpnet import _ProfileLogitScaling
from bpnetlite.chrombpnet import _Log, _Exp

from tangermeme.ersatz import dinucleotide_shuffle
from tangermeme.deep_lift_shap import deep_lift_shap as t_deep_lift_shap
from tangermeme.deep_lift_shap import _nonlinear


def deep_lift_shap(model, X, args=None, target=0,  batch_size=32,
	references=dinucleotide_shuffle, n_shuffles=20, return_references=False, 
	hypothetical=False, warning_threshold=0.001, additional_nonlinear_ops=None,
	print_convergence_deltas=False, raw_outputs=False, device='cuda', 
	random_state=None, verbose=False):
	"""A wrapper that registers Chrom/BPNet's custom non-linearities.

	This function is just a wrapper for tangermeme's deep_lift_shap function
	except that it automatically registers the layers that are necessary for
	using BPNet models. Specifically, it registers a scaling that is necessary
	for calculating the profile attributions and also registers the logsumexp
	operation for counts when using the full ChromBPNet model.

	Other than automatically registering the non-linearities, this wrapper does
	not modify the tangermeme outputs or alter the inputs in any way. It is
	simply for convenience so you do not need to reach into bpnet-lite's
	internals each time you want to calculate attributions.


	Parameters
	----------
	model: torch.nn.Module
		A PyTorch model to use for making predictions. These models can take in
		any number of inputs and make any number of outputs. The additional
		inputs must be specified in the `args` parameter.

	X: torch.tensor, shape=(-1, len(alphabet), length)
		A set of one-hot encoded sequences to calculate attribution values
		for. 

	args: tuple or None, optional
		An optional set of additional arguments to pass into the model. If
		provided, each element in the tuple or list is one input to the model
		and the element must be formatted to be the same batch size as `X`. If
		None, no additional arguments are passed into the forward function.
		Default is None.

	target: int, optional
		The output of the model to calculate gradients/attributions for. This
		will index the last dimension of the predictions. Default is 0.

	batch_size: int, optional
		The number of sequence-reference pairs to pass through DeepLiftShap at
		a time. Importantly, this is not the number of elements in `X` that
		are processed simultaneously (alongside ALL their references) but the
		total number of `X`-`reference` pairs that are processed. This means
		that if you are in a memory-limited setting where you cannot process
		all references for even a single sequence simultaneously that the
		work is broken down into doing only a few references at a time. Default
		is 32.

	references: func or torch.Tensor, optional
		If a function is passed in, this function is applied to each sequence
		with the provided random state and number of shuffles. This function
		should serve to transform a sequence into some form of signal-null
		background, such as by shuffling it. If a torch.Tensor is passed in,
		that tensor must have shape `(len(X), n_shuffles, *X.shape[1:])`, in
		that for each sequence a number of shuffles are provided. Default is
		the function `dinucleotide_shuffle`. 

	n_shuffles: int, optional
		The number of shuffles to use if a function is given for `references`.
		If a torch.Tensor is provided, this number is ignored. Default is 20.

	return_references: bool, optional
		Whether to return the references that were generated during this
		process. Only use if `references` is not a torch.Tensor. Default is 
		False. 

	hypothetical: bool, optional
		Whether to return attributions for all possible characters at each
		position or only for the character that is actually at the sequence.
		Practically, whether to return the returned attributions from captum
		with the one-hot encoded sequence. Default is False.

	warning_threshold: float, optional
		A threshold on the convergence delta that will always raise a warning
		if the delta is larger than it. Normal deltas are in the range of
		1e-6 to 1e-8. Note that convergence deltas are calculated on the
		gradients prior to the aggr_func being applied to them. Default 
		is 0.001. 

	additional_nonlinear_ops: dict or None, optional
		If additional nonlinear ops need to be added to the dictionary of
		operations that can be handled by DeepLIFT/SHAP, pass a dictionary here
		where the keys are class types and the values are the name of the
		function that handle that sort of class. Make sure that the signature
		matches those of `_nonlinear` and `_maxpool` above. This can also be
		used to overwrite the hard-coded operations by passing in a dictionary
		with overlapping key names. If None, do not add any additional 
		operations. Default is None.

	print_convergence_deltas: bool, optional
		Whether to print the convergence deltas for each example when using
		DeepLiftShap. Default is False.

	raw_outputs: bool, optional
		Whether to return the raw outputs from the method -- in this case,
		the multipliers for each example-reference pair -- or the processed
		attribution values. Default is False.

	device: str or torch.device, optional
		The device to move the model and batches to when making predictions. If
		set to 'cuda' without a GPU, this function will crash and must be set
		to 'cpu'. Default is 'cuda'. 

	random_state: int or None or numpy.random.RandomState, optional
		The random seed to use to ensure determinism. If None, the
		process is not deterministic. Default is None. 

	verbose: bool, optional
		Whether to display a progress bar. Default is False.


	Returns
	-------
	attributions: torch.tensor
		If `raw_outputs=False` (default), the attribution values with shape
		equal to `X`. If `raw_outputs=True`, the multipliers for each example-
		reference pair with shape equal to `(X.shape[0], n_shuffles, X.shape[1],
		X.shape[2])`. 

	references: torch.tensor, optional
		The references used for each input sequence, with the shape
		(n_input_sequences, n_shuffles, 4, length). Only returned if
		`return_references = True`. 
	"""

	return t_deep_lift_shap(model=model, X=X, args=args, target=target, 
		batch_size=batch_size, references=references, n_shuffles=n_shuffles, 
		return_references=return_references, hypothetical=hypothetical, 
		warning_threshold=warning_threshold,
		additional_nonlinear_ops={
			_ProfileLogitScaling: _nonlinear,
			_Log: _nonlinear,
			_Exp: _nonlinear
		},
		print_convergence_deltas=print_convergence_deltas, 
		raw_outputs=raw_outputs, device=device, random_state=random_state, 
		verbose=verbose)



# io.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>
# Code adapted from Alex Tseng, Avanti Shrikumar, and Ziga Avsec

import numpy
import torch

from tqdm import tqdm
from tangermeme.io import extract_loci

import time


class PeakNegativeSampler(torch.utils.data.Dataset):
	"""A data generator mimicking the BPNet data loading procedure.
	
	Here, a set of peaks and negatives are separately loaded. These sets can be
	any size. From these sets, batches of given size are sampled that are a
	mixture of peaks and negatives. The peaks are sampled by randomly iterating
	over the entire set such that one epoch means one pass over the entire data
	set. The negatives are sampled by randomly choosing regions from the
	negatives at the given ratio to peaks, without considering whether they have
	been selected before.
	
	Because peaks and negatives are provided as separate tensors, different
	jittering can be used on them. This means that, for instance, jittering can
	be used on peaks but not used on negatives.
	
	In the documentation below, `mj` = max_jitter.
	
	Note that, although the data is passed in as PyTorch tensors, they are saved
	as numpy arrays for faster slicing during training.
	
	
	Parameters
	----------
	peak_sequences: torch.tensor, shape=(n_peaks, 4, in_window+2*mj)
		A tensor of peak sequences that are one-hot encoded. See above for the
		connection between the length here and jitter.

	peak_signals: torch.tensor, shape=(n_peaks, t, out_window+2*mj)
		A tensor of signals to predict, usually base-pair resolution integer counts.
		This should have `t` tasks, which is usually 2 if predicting stranded outputs
		and 1 if predicting unstranded outputs. See above for the connection between
		the length here and jitter.

	peak_controls: torch.tensor, shape=(n, t, out_window+2*mj) or None, optional
		A tenso of the control signal to take as input, usually base-pair counts, for `n`
		examples with `t` strands and output length `out_window`. If
		None, does not return controls.

	negative_sequences: torch.tensor, shape=(n, 4, in_window+2*max_jitter)
		A one-hot encoded tensor of `n` example sequences, each of input 
		length `in_window`. See description above for connection with jitter.

	negative_signals: torch.tensor, shape=(n, t, out_window+2*max_jitter)
		The signals to predict, usually counts, for `n` examples with
		`t` output tasks (usually 2 if stranded, 1 otherwise), each of 
		output length `out_window`. See description above for connection 
		with jitter.

	negative_controls: torch.tensor, shape=(n, t, out_window+2*max_jitter) or None, optional
		The control signal to take as input, usually counts, for `n`
		examples with `t` strands and output length `out_window`. If
		None, does not return controls.
	
	p: torch.tensor or None, shape=(n,) 
		A vector of probabilities that sum to 1 containing the sampling probability
		of each sequence. If None, use a uniform distribution.

	in_window: int, optional
		The input window size. Default is 2114.

	out_window: int, optional
		The output window size. Default is 1000.

	max_jitter: int, optional
		The maximum amount of jitter to add, in either direction, to the
		midpoints that are passed in. Default is 0.

	reverse_complement: bool, optional
		Whether to reverse complement-augment half of the data. Default is False.

	random_state: int or None, optional
		Whether to use a deterministic seed or not.
	"""
	
	def __init__(self, peak_sequences, peak_signals, negative_sequences, 
		negative_signals, peak_controls=None, negative_controls=None, 
		negative_ratio=0.1, in_window=2114, out_window=1000, max_jitter=0, 
		reverse_complement=False, shuffle=True, random_state=None):
		self.peak_sequences = peak_sequences.numpy(force=True)
		self.peak_signals = peak_signals.numpy(force=True)
		self.peak_controls = peak_controls.numpy(force=True)
		self.n_peaks = len(self.peak_sequences)
		
		self.negative_sequences = negative_sequences.numpy(force=True)
		self.negative_signals = negative_signals.numpy(force=True)
		self.negative_controls = negative_controls.numpy(force=True)
		self.n_negatives = len(self.negative_sequences)

		self.negative_ratio = negative_ratio
		self.negative_likelihood = 1 / (1 + 1 / negative_ratio)

		self.in_window = in_window
		self.out_window = out_window
		self.max_jitter = max_jitter
		self.reverse_complement = reverse_complement
		self.shuffle = shuffle

		self.random_state = numpy.random.RandomState(random_state)
		self.n_peaks_seen = 0
		self.peak_ordering = None

	def __len__(self):
		return self.n_peaks + int(self.n_peaks * self.negative_ratio)

	def __getitem__(self, idx):
		if idx == 0:
			self.peak_ordering = numpy.arange(self.n_peaks)
			if self.shuffle:
				self.random_state.shuffle(self.peak_ordering)


		if self.random_state.uniform() >= self.negative_likelihood:
			idx = self.peak_ordering[self.n_peaks_seen % self.n_peaks]
			jitter = self.random_state.randint(self.max_jitter*2)
			label = 1
 
			X, y, X_ctl = self.peak_sequences, self.peak_signals, self.peak_controls
			self.n_peaks_seen += 1

		else:
			idx = self.random_state.randint(self.n_negatives)
			jitter = 0
			label = 0

			X, y, X_ctl = (self.negative_sequences, self.negative_signals, 
				self.negative_controls)


		Xi = torch.from_numpy(X[idx][:, jitter:jitter+self.in_window])
		yi = torch.from_numpy(y[idx][:, jitter:jitter+self.out_window])
		if self.peak_controls is not None:
			Xi_ctl = torch.from_numpy(X_ctl[idx][:, jitter:jitter+self.in_window])


		if self.reverse_complement and self.random_state.randint(2) == 1:
			Xi = torch.flip(Xi, [0, 1])
			yi = torch.flip(yi, [0, 1])
			
			if self.peak_controls is not None:
				Xi_ctl = torch.flip(Xi_ctl, [0, 1])


		if self.peak_controls is not None:
			return Xi, Xi_ctl, yi, label
		
		return Xi, yi, label


class DataGenerator(torch.utils.data.Dataset):
	"""A data generator for BPNet inputs.

	This generator takes in an extracted set of sequences, output signals,
	and control signals, and will return a single element with random
	jitter and reverse-complement augmentation applied. Jitter is implemented
	efficiently by taking in data that is wider than the in/out windows by
	two times the maximum jitter and windows are extracted from that.
	Essentially, if an input window is 1000 and the maximum jitter is 128, one
	would pass in data with a length of 1256 and a length 1000 window would be
	extracted starting between position 0 and 256. This  generator must be 
	wrapped by a PyTorch generator object.

	Parameters
	----------
	sequences: torch.tensor, shape=(n, 4, in_window+2*max_jitter)
		A one-hot encoded tensor of `n` example sequences, each of input 
		length `in_window`. See description above for connection with jitter.

	signals: torch.tensor, shape=(n, t, out_window+2*max_jitter)
		The signals to predict, usually counts, for `n` examples with
		`t` output tasks (usually 2 if stranded, 1 otherwise), each of 
		output length `out_window`. See description above for connection 
		with jitter.

	controls: torch.tensor, shape=(n, t, out_window+2*max_jitter) or None, optional
		The control signal to take as input, usually counts, for `n`
		examples with `t` strands and output length `out_window`. If
		None, does not return controls.
	
	p: torch.tensor or None, shape=(n,) 
		A vector of probabilities that sum to 1 containing the sampling probability
		of each sequence. If None, use a uniform distribution.

	in_window: int, optional
		The input window size. Default is 2114.

	out_window: int, optional
		The output window size. Default is 1000.

	max_jitter: int, optional
		The maximum amount of jitter to add, in either direction, to the
		midpoints that are passed in. Default is 0.

	reverse_complement: bool, optional
		Whether to reverse complement-augment half of the data. Default is False.

	random_state: int or None, optional
		Whether to use a deterministic seed or not.
	"""

	def __init__(self, sequences, signals, controls=None, p=None, in_window=2114, 
		out_window=1000, max_jitter=0, reverse_complement=False, 
		random_state=None):
		self.p = p
		self.in_window = in_window
		self.out_window = out_window
		self.max_jitter = max_jitter
		
		self.reverse_complement = reverse_complement
		self.random_state = numpy.random.RandomState(random_state)

		self.signals = signals
		self.controls = controls
		self.sequences = sequences
		
		self.random_idxs = None
		self.n_random = 1000000

	def __len__(self):
		return len(self.sequences)

	def __getitem__(self, idx):
		if idx % self.n_random == 0:
			self.random_idxs = self.random_state.choice(len(self), p=self.p, 
				size=self.n_random)

		i = self.random_idxs[idx % self.n_random]
		j = 0 if self.max_jitter == 0 else self.random_state.randint(
			self.max_jitter*2) 

		X = self.sequences[i][:, j:j+self.in_window]
		y = self.signals[i][:, j:j+self.out_window]

		if self.controls is not None:
			X_ctl = self.controls[i][:, j:j+self.in_window]

		if self.reverse_complement and self.random_state.choice(2) == 1:
			X = torch.flip(X, [0, 1])
			y = torch.flip(y, [0, 1])

			if self.controls is not None:
				X_ctl = torch.flip(X_ctl, [0, 1])

		if self.controls is not None:
			return X, X_ctl, y

		return X, y


def PeakGenerator(peaks, negatives, sequences, signals, controls=None,
	chroms=None, in_window=2114, out_window=1000, max_jitter=128, 
	negative_ratio=0.1, reverse_complement=True, shuffle=True, min_counts=None, 
	max_counts=None, summits=False, exclusion_lists=None, random_state=None, 
	pin_memory=True, num_workers=0, batch_size=32, verbose=False):
	"""This is a constructor function that handles all IO.

	This function will extract signal from all signal and control files,
	pass that into a DataGenerator, and wrap that using a PyTorch data
	loader. This is the only function that needs to be used.


	Parameters
	----------
	peaks: str or pandas.DataFrame or list/tuple of such
		A BED-formatted file containing peak coordinates. This can be either
		the string path to the BED file or a pandas DataFrame object containing
		three columns: chrom, start, and end. Alternatively, this can be a list
		of such objects whose coordinates will be interleaved.

	negatives: str or pandas.DataFrame or list/tuple of such
		A BED-formatted file containing negative coordinates. This can be either
		the string path to the BED file or a pandas DataFrame object containing
		three columns: chrom, start, and end. Alternatively, this can be a list
		of such objects whose coordinates will be interleaved.

	sequences: str or dictionary
		Either the path to a fasta file to read from or a dictionary where the
		keys are the unique set of chromosoms and the values are one-hot
		encoded sequences as numpy arrays or memory maps.

	signals: list of strs or list of dictionaries
		A list of filepaths to bigwig files, where each filepath will be read
		using pyBigWig, or a list of dictionaries where the keys are the same
		set of unique chromosomes and the values are numpy arrays or memory
		maps.

	controls: list of strs or list of dictionaries or None, optional
		A list of filepaths to bigwig files, where each filepath will be read
		using pyBigWig, or a list of dictionaries where the keys are the same
		set of unique chromosomes and the values are numpy arrays or memory
		maps. If None, no control tensor is returned. Default is None. 

	chroms: list or None, optional
		A set of chromosomes to extact loci from. Loci in other chromosomes
		in the locus file are ignored. If None, all loci are used. Default is
		None.

	in_window: int, optional
		The input window size. Default is 2114.

	out_window: int, optional
		The output window size. Default is 1000.

	max_jitter: int, optional
		The maximum amount of jitter to add, in either direction, to the
		midpoints that are passed in. Default is 128.

	negative_ratio: float, optional
		The ratio of negatives compared to peaks in each batch. A value of 1 means
		that each batch is balanced, and a value of 10 means that there would be 10
		negatives for each positive. Note that this is independent of the number of
		peaks and negatives provided. Even if the `peaks` input has 10x the number
		of coordinates as the `negatives` one, if the ratio is 1 each batch during
		training will be balanced (on average).

	reverse_complement: bool, optional
		Whether to reverse complement-augment half of the data. Default is True.

	shuffle: bool, optional
		Whether to randomly sample peaks, if True, or to proceed sequentially
		through them, if False. Negatives are always randomly sampled. Default
		is True.

	min_counts: float or None, optional
		The minimum number of counts, summed across the length of each example
		and across all tasks, needed to be kept. If None, no minimum. Default 
		is None.

	max_counts: float or None, optional
		The maximum number of counts, summed across the length of each example
		and across all tasks, needed to be kept. If None, no maximum. Default 
		is None.  

	summits: bool, optional
		Whether to return a region centered around the summit instead of the center
		between the start and end. If True, it will add the 10th column (index 9)
		to the start to get the center of the window, and so the data must be in 
		narrowPeak format.

	exclusion_lists: list or None, optional
		A list of strings of filenames to BED-formatted files containing exclusion
		lists, i.e., regions where overlapping loci should be filtered out. If None,
		no filtering is performed based on exclusion zones. Default is None.

	random_state: int or None, optional
		Whether to use a deterministic seed or not.

	pin_memory: bool, optional
		Whether to pin page memory to make data loading onto a GPU easier.
		Default is True.

	num_workers: int, optional
		The number of processes fetching data at a time to feed into a model.
		If 0, data is fetched from the main process. Default is 0.

	batch_size: int, optional
		The number of data elements per batch. Default is 32.
	
	verbose: bool, optional
		Whether to display a progress bar while loading. Default is False.

	Returns
	-------
	X: torch.utils.data.DataLoader
		A PyTorch DataLoader wrapped DataGenerator object.
	"""

	X_peaks = extract_loci(loci=peaks, sequences=sequences, 
		signals=signals, in_signals=controls, chroms=chroms, in_window=in_window, 
		out_window=out_window, max_jitter=max_jitter, min_counts=min_counts,
		max_counts=max_counts, summits=summits, exclusion_lists=exclusion_lists,
		ignore=list('QWERYUIOPSDFHJKLZXVBNM'), return_mask=True, verbose=verbose)

	loci_counts = X_peaks[1].sum(dim=(1, 2))
	
	outlier_threshold = torch.quantile(X_peaks[1].sum(dim=(1, 2)), 0.99) * 1.2
	outlier_idxs = loci_counts > outlier_threshold

	X_bg = extract_loci(loci=negatives, sequences=sequences, 
		signals=signals, in_signals=controls, chroms=chroms, in_window=in_window, 
		out_window=out_window, max_jitter=0, min_counts=min_counts,
		max_counts=max_counts, summits=False, exclusion_lists=exclusion_lists,
		ignore=list('QWERYUIOPSDFHJKLZXVBNM'), return_mask=True, verbose=verbose)

	if verbose:
		n_filtered_peaks = len(X_peaks[-1]) - X_peaks[-1].sum() + outlier_idxs.sum()
		n_filtered_negatives = len(X_bg[-1]) - X_bg[-1].sum()
		
		print("\nFiltered Peaks: {}".format(n_filtered_peaks))
		print("Filtered Negatives: {}".format(n_filtered_negatives))
	
	###

	X_gen = PeakNegativeSampler(
		peak_sequences=X_peaks[0][~outlier_idxs],
		peak_signals=X_peaks[1][~outlier_idxs],
		peak_controls=None if controls is None else X_peaks[2][~outlier_idxs],
		negative_sequences=X_bg[0],
		negative_signals=X_bg[1],
		negative_controls=None if controls is None else X_bg[2],
		negative_ratio=negative_ratio,
		in_window=in_window,
		out_window=out_window,
		max_jitter=max_jitter,
		reverse_complement=reverse_complement,
		shuffle=shuffle,
		random_state=random_state
	)

	X_gen = torch.utils.data.DataLoader(X_gen, pin_memory=pin_memory,
		num_workers=num_workers, batch_size=batch_size) 

	return X_gen



# marginalize.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import os
import numpy
import torch
import pandas
import seaborn

from bpnetlite.bpnet import _ProfileLogitScaling
from bpnetlite.bpnet import CountWrapper
from bpnetlite.bpnet import ProfileWrapper
from bpnetlite.chrombpnet import _Log, _Exp

from tangermeme.io import read_meme
from tangermeme.io import one_hot_encode

from tangermeme.marginalize import marginalize
from tangermeme.deep_lift_shap import deep_lift_shap
from tangermeme.deep_lift_shap import _nonlinear

from tangermeme.plot import plot_logo

from tqdm import tqdm

import matplotlib.pyplot as plt


def path_to_image_html(path):
	return '<img src="' + path + '" width="240" >'


def _plot_profiles(y, ylim, color, path, figsize=(10,3)):
	plt.figure(figsize=figsize)
	plt.plot(y, color=color)

	seaborn.despine()
	plt.xlim(0, y.shape[0])
	plt.xlabel("Relative Coordinate", fontsize=12)
	plt.ylabel("Predicted Signal", fontsize=12)
	plt.ylim(*ylim)
	plt.yticks(fontsize=12)
	plt.savefig(path)
	plt.close()


def _plot_counts(y_before, y_after, xlim, ylim, color, path, figsize=(10,3)):
	zmax = max(xlim[1], ylim[1])
	zmin = min(xlim[0], ylim[0])

	plt.figure(figsize=figsize)
	plt.plot([zmin, zmax], [zmin, zmax], color='0.5')
	plt.scatter(y_before, y_after, color=color, s=5)

	seaborn.despine()
	plt.xlabel("Pred Counts Before", fontsize=12)
	plt.ylabel("Pred Counts After", fontsize=12)
	plt.xlim(*xlim)
	plt.ylim(*ylim)
	plt.yticks(fontsize=12)
	plt.savefig(path)
	plt.close()


def _plot_attributions(y, ylim, path, figsize=(10,3), **kwargs):
	fig = plt.figure(figsize=figsize)
	ax = fig.add_subplot(111)

	plot_logo(y.T, ax=ax)
	
	plt.yticks(fontsize=12)
	plt.ylim(*ylim)
	plt.savefig(path)
	plt.close()


def marginalization_report(model, motifs, X, output_dir, batch_size=64,  
	attributions=True, minimal=False, device='cuda', verbose=False):
	"""Create an HTML report showing the impact of each motif.

	Take in a predictive model, a MEME file of motifs, and a set of sequences,
	and run a marginalization experiment on each motif. Store the images
	to the output directory, and then create an HTML report that puts together
	these images.


	Parameters
	----------
	model: bpnetlite.bpnet.BPNet or bpnetlite.chrombpnet.ChromBPNet
		A BPNet- or ChromBPNet-style model that outputs predictions for a
		profile head and a count head.

	motif: str
		A motif to insert into the middle of the background sequences.

	X: torch.Tensor, shape=(n, 4, 2114)
		A one-hot encoded set of n sequences to run through the model.

	output_dir: str
		The folder name to put all the images that are generated.

	batch_size: int, optional
		The number of examples to run at a time in each batch. Default is 64.

	attributions: bool, optional
		Whether to calculate attributions as well as calculating predictions
		when doing marginalizations. Because calculating attributions is the
		most time-intensive aspect, setting this parameter to False can save
		time. Default is True

	minimal: bool, optional
		Whether to produce a minimal report, which shows the differences in
		outputs, or the full report, which shows the results before and after
		insertion as well as the differences. Potentially useful for debugging.
		Default is False.
	
	device: str, optional
		The device to run operations on. Usually is either 'cuda' or 'cpu'.
		Default is 'cuda'.

	verbose: bool, optional
		Whether to print a progress bar as motifs are marginalized over. Default
		is False.
	"""

	motifs = list(read_meme(motifs).items())

	if not os.path.isdir(output_dir):
		os.mkdir(output_dir)

	results = {
		'name': [],
		'sequence': [],
		'profile (before)': [],
		'profile (after)': [],
		'profile (diff)': [],
		'counts': [],
		'profile attributions (before)': [],
		'profile attributions (after)': [],
		'profile attributions (diff)': [],
		'counts attributions (before)': [],
		'counts attributions (after)': [],
		'counts attributions (diff)': []
	}


	p, c, a = 'profile', 'counts', ' attributions'
	pb, pa, pd = '.profile.before', '.profile.after', '.profile.diff'
	pab, paa, pad = '.pattr.before', '.pattr.after', '.pattr.diff'
	cab, caa, cad = '.cattr.before', '.cattr.after', '.cattr.diff'

	mid, w = X.shape[-1] // 2, 15
	s, e = mid - w, mid + w

	p_before, p_after, p_diff = [], [], []
	c_before, c_after, c_diff = [], [], []
	ap_before, ap_after, ap_diff = [], [], []
	ac_before, ac_after, ac_diff = [], [], []

	for i, (name, pwm) in tqdm(enumerate(motifs), disable=not verbose):
		motif = ''.join(numpy.array(['A', 'C', 'G', 'T'])[pwm.argmax(axis=0)])

		(y_profile_before, y_counts_before), (y_profile_after, 
			y_counts_after) = marginalize(model, X, motif, 
			batch_size=batch_size, device=device)

		y_profile_before = torch.nn.functional.softmax(y_profile_before, dim=-1)
		y_profile_after = torch.nn.functional.softmax(y_profile_after, dim=-1)

		p_before.append(y_profile_before.mean(axis=0).T)
		p_after.append(y_profile_after.mean(axis=0).T)
		p_diff_ = (y_profile_after - y_profile_before).mean(axis=0).T
		p_diff.append(p_diff_)

		c_before.append(y_counts_before)
		c_after.append(y_counts_after)

		if attributions:
			Xp_attr_before, Xp_attr_after = marginalize(ProfileWrapper(
				model.double()), X.double(), motif, func=deep_lift_shap, 
				additional_nonlinear_ops={
					_ProfileLogitScaling: _nonlinear,
					_Log: _nonlinear,
					_Exp: _nonlinear
				}, 
				n_shuffles=1, batch_size=batch_size, device=device)

			Xc_attr_before, Xc_attr_after = marginalize(
				CountWrapper(model).double(), X.double(), motif, 
				func=deep_lift_shap, n_shuffles=1,
				additional_nonlinear_ops={
					_ProfileLogitScaling: _nonlinear,
					_Log: _nonlinear,
					_Exp: _nonlinear
				}, 
				batch_size=batch_size,
				device=device)

			ap_before.append(Xp_attr_before.mean(axis=0)[:, s:e].T)
			ap_after.append(Xp_attr_after.mean(axis=0)[:, s:e].T)
			ap_diff_ = (Xp_attr_after - Xp_attr_before).mean(axis=0)[:, s:e].T
			ap_diff.append(ap_diff_)

			ac_before.append(Xc_attr_before.mean(axis=0)[:, s:e].T)
			ac_after.append(Xc_attr_after.mean(axis=0)[:, s:e].T)
			ac_diff_ = (Xc_attr_after - Xc_attr_before).mean(axis=0)[:, s:e].T
			ac_diff.append(ac_diff_)

	multistack = lambda args: [torch.stack(a).numpy() for a in args]
	p_before, p_after, p_diff = multistack([p_before, p_after, p_diff])
	c_before, c_after = multistack([c_before, c_after])

	multilims = lambda args: [(x.min() * 0.95, x.max() * 1.05) for x in args] 
	pb_lim, pa_lim, pd_lim = multilims([p_before, p_after, p_diff])
	cb_lim, ca_lim = multilims([c_before, c_after])

	if attributions:
		apb, apa, apd = multistack([ap_before, ap_after, ap_diff])
		acb, aca, acd = multistack([ac_before, ac_after, ac_diff])

		apb_lim, apa_lim, apd_lim = multilims([apb, apa, apd])
		acb_lim, aca_lim, acd_lim = multilims([acb, aca, acd])


	idxs = (c_after - c_before).mean(axis=1)[:, 0].argsort()[::-1]
	for i, idx in enumerate(idxs):
		name, pwm = motifs[idx]
		name = name.replace(" ", "_").replace("/", "-")
		oname = output_dir + name
		motif = ''.join(numpy.array(['A', 'C', 'G', 'T'])[pwm.argmax(axis=0)])

		if not minimal:
			_plot_profiles(p_before[idx], pb_lim, color='0.5', 
				path=oname + pb + ".png")
			_plot_profiles(p_after[idx], pa_lim, color='0.5', 
				path=oname + pa + ".png")

		_plot_profiles(p_diff[idx], pd_lim, color='c', 
			path=oname + pd + ".png")
		_plot_counts(c_before[idx], c_after[idx], cb_lim, ca_lim, color='m', 
			path=oname + ".counts.png")

		if attributions:
			if not minimal:
				_plot_attributions(apb[idx], apb_lim, oname + pab + ".png")
				_plot_attributions(apa[idx], apa_lim, oname + paa + ".png")

				_plot_attributions(acb[idx], acb_lim, oname + cab + ".png")
				_plot_attributions(aca[idx], aca_lim, oname + caa + ".png")

			_plot_attributions(apd[idx], apd_lim, oname + pad + ".png")
			_plot_attributions(acd[idx], acd_lim, oname + cad + ".png")

		motif_ = motif[:25] + ('...' if len(motif) > 25 else '')

		results['name'].append(name)
		results['sequence'].append(motif_)
		results[p + ' (before)'].append(name + pb + ".png")
		results[p + ' (after)'].append(name + pa + ".png")
		results[p + ' (diff)'].append(name + pd + ".png")
		results[c].append(name + ".counts.png")

		if attributions:
			results[p + a + ' (before)'].append(name + pab + ".png")
			results[p + a + ' (after)'].append(name + paa + ".png")
			results[p + a + ' (diff)'].append(name + pad + ".png")
			results[c + a + ' (before)'].append(name + cab + ".png")
			results[c + a + ' (after)'].append(name + caa + ".png")
			results[c + a + ' (diff)'].append(name + cad + ".png")

	if not attributions:
		del results[p + a + ' (before)']
		del results[p + a + ' (after)']
		del results[p + a + ' (diff)']
		del results[c + a + ' (before)']
		del results[c + a + ' (after)']
		del results[c + a + ' (diff)']

	formatters = {name: path_to_image_html for name in results.keys() 
		if name not in ('name', 'sequence')}

	results_df = pandas.DataFrame(results)
	if minimal:
		if attributions:
			results_df = results_df[['name', 'sequence', 'profile (diff)', 
				'counts', 'profile attributions (diff)', 
				'counts attributions (diff)']]
		else:
			results_df = results_df[['name', 'sequence', 'profile (diff)', 
				'counts']]
			
	results_df.to_html(open('{}/marginalization.html'.format(output_dir), 'w'),
		escape=False, formatters=formatters, index=False)



# performance.py
# Authors: Alex Tseng <amtseng@stanford.edu>
#          Jacob Schreiber <jmschreiber91@gmail.com>

"""
This module contains performance measures that are used to evaluate the
model but are not explicitly used as losses for optimization.

IMPORTANT: MANY OF THESE FUNCTIONS ASSUME THE INPUTS TO BE PREDICTED LOG
PROBABILITIES AND TRUE COUNTS. THE FIRST ARGUMENT MUST BE IN LOG SPACE
AND THE SECOND ARGUMENT MUST BE IN COUNT SPACE FOR THESE FUNCTIONS.
"""

import torch

from .losses import MNLLLoss
from .losses import log1pMSELoss


def smooth_gaussian1d(x, kernel_sigma, kernel_width):
	"""Smooth a signal along the sequence length axis.

	This function is a replacement for the scipy.ndimage.gaussian1d
	function that works on PyTorch tensors. It applies a Gaussian kernel
	to each position which is equivalent to applying a convolution across
	the sequence with weights equal to that of a Gaussian distribution.
	Each sequence, and each channel within the sequence, is smoothed
	independently.


	Parameters
	----------
	x: torch.tensor, shape=(n_sequences, n_channels, seq_len)
		A tensor to smooth along the last axis. n_channels must be at
		least 1.

	kernel_sigma: float
		The standard deviation of the Gaussian to be applied.

	kernel_width: int
		The width of the kernel to be applied.


	Returns
	-------
	x_smooth: torch.tensor, shape=(n_sequences, n_channels, seq_len)
		The smoothed tensor.
	"""

	meshgrid = torch.arange(kernel_width, dtype=torch.float32,
		device=x.device)

	mean = (kernel_width - 1.) / 2.
	kernel = torch.exp(-0.5 * ((meshgrid - mean) / kernel_sigma) ** 2.0)
	kernel = kernel / torch.sum(kernel)
	kernel = kernel.reshape(1, 1, kernel_width).repeat(x.shape[1], 1, 1)
	return torch.nn.functional.conv1d(x, weight=kernel, groups=x.shape[1], 
		padding='same')


def batched_smoothed_function(logps, true_counts, f, smooth_predictions=False, 
	smooth_true=False, kernel_sigma=7, kernel_width=81, 
	exponentiate_logps=False, batch_size=200):
	"""Batch a calculation with optional smoothing.

	Given a set of predicted and true values, apply some function to them in
	a batched manner and store the results. Optionally, either the true values
	or the predicted ones can be smoothed.


	Parameters
	----------
	logps: torch.tensor
		A tensor of the predicted log probability values.

	true_counts: torch.tensor
		A tensor of the true values, usually integer counts.

	f: function
		A function to be applied to the predicted and true values.

	smooth_predictions: bool, optional
		Whether to apply a Gaussian filter to the predictions. Default is 
		False.

	smooth_true: bool, optional
		Whether to apply a Gaussian filter to the true values. Default is
		False.

	kernel_sigma: float, optional
		The standard deviation of the Gaussian to be applied. Default is 7.

	kernel_width: int, optional
		The width of the kernel to be applied. Default is 81.

	exponentiate_logps: bool, optional
		Whether to exponentiate each batch of log probabilities. Default is
		False.

	batch_size: int, optional
		The number of examples in each batch to evaluate at a time. Default
		is 200.


	Returns
	-------
	results: torch.tensor
		The results of applying the function to the tensor.
	"""

	n = logps.shape[0]
	results = torch.empty(*logps.shape[:2])

	for start in range(0, n, batch_size):
		end = start + batch_size

		logps_ = logps[start:end]
		true_counts_ = true_counts[start:end]

		if smooth_predictions:
			logps_ = torch.exp(logps_)
			logps_ = smooth_gaussian1d(logps_, kernel_sigma, kernel_width)

			if exponentiate_logps == False:
				logps_ = torch.log(logps_)
		else:
			if exponentiate_logps:
				logps_ = torch.exp(logps_)

		if smooth_true:
			true_counts_ = smooth_gaussian1d(true_counts_, kernel_sigma, kernel_width)

		results[start:end] = f(logps_, true_counts_) 

	return results


def _kl_divergence(probs1, probs2):
	"""The KL-divergence between two tensors across the last axis.
	
	Computes the KL divergence in the last dimension of `probs1` and `probs2`
	as KL(P1 || P2). `probs1` and `probs2` must be the same shape. For example,
	if they are both A x B x L arrays, then the KL divergence of corresponding
	L-arrays will be computed and returned in an A x B array. Does not
	renormalize the arrays. If probs2[i] is 0, that value contributes 0.


	Parameters
	----------
	logps: torch.Tensor
		A tensor of probabilities. This must not be log probabilities.

	true_counts: torch.Tensor
		A tensor of probabilities. This must not be log probabilities.


	Returns
	-------
	kl: torch.Tensor
		The KL divergence for each element.
	"""

	idxs = ((probs1 != 0) & (probs2 != 0))
	quot_ = torch.divide(probs1, probs2)

	quot = torch.ones_like(probs1)
	quot[idxs] = quot_[idxs]
	return torch.sum(probs1 * torch.log(quot), dim=-1)


def jensen_shannon_distance(logps, true_counts):
	"""The Jensen-Shannon distance between two tensors across the last axis.

	Computes the Jensen-Shannon distance in the last dimension of `logps` and
	`true_counts`. These two tensors must be the same shape. For example, if they
	are both A x B x L arrays, then the KL divergence of corresponding L-arrays
	will be computed and returned in an A x B array. This will renormalize the
	arrays so that each subarray sums to 1. If the sum of a subarray is 0, then
	the resulting JSD will be NaN.


	Parameters
	----------
	logps: torch.Tensor
		A tensor of log probabilities or logits. This must be in log space.

	true_counts: torch.Tensor
		A tensor of true integer counts at each position.


	Returns
	-------
	jsd: torch.Tensor
		The Jensen-Shannon divergence for each element.
	"""
	# Renormalize both distributions, and if the sum is NaN, put NaNs all around

	probs1 = torch.exp(logps)
	probs1_sum = torch.sum(probs1, dim=-1, keepdims=True)
	probs1 = torch.divide(probs1, probs1_sum, out=torch.zeros_like(probs1))

	probs2_sum = torch.sum(true_counts, dim=-1, keepdims=True)
	probs2 = torch.divide(true_counts, probs2_sum, out=torch.zeros_like(true_counts))
	probs2 = probs2.type(probs1.dtype)

	mid = 0.5 * (probs1 + probs2)
	return 0.5 * (_kl_divergence(probs1, mid) + _kl_divergence(probs2, mid))


def pearson_corr(arr1, arr2):
	"""The Pearson correlation between two tensors across the last axis.

	Computes the Pearson correlation in the last dimension of `arr1` and `arr2`.
	`arr1` and `arr2` must be the same shape. For example, if they are both
	A x B x L arrays, then the correlation of corresponding L-arrays will be
	computed and returned in an A x B array.


	Parameters
	----------
	arr1: torch.Tensor
		One of the tensor to correlate.

	arr2: torch.Tensor
		The other tensor to correlation.


	Returns
	-------
	correlation: torch.Tensor
		The correlation for each element where `n` is arr1.shape[-1].
	"""

	mean1 = torch.mean(arr1, axis=-1).unsqueeze(-1)
	mean2 = torch.mean(arr2, axis=-1).unsqueeze(-1)
	dev1, dev2 = arr1 - mean1, arr2 - mean2

	sqdev1, sqdev2 = torch.square(dev1), torch.square(dev2)
	numer = torch.sum(dev1 * dev2, axis=-1)  # Covariance
	var1, var2 = torch.sum(sqdev1, axis=-1), torch.sum(sqdev2, axis=-1)  # Variances
	denom = torch.sqrt(var1 * var2)

	# Divide numerator by denominator, but use 0 where the denominator is 0
	correlation = torch.zeros_like(numer)
	correlation[denom != 0] = numer[denom != 0] / denom[denom != 0]
	return correlation


def spearman_corr(arr1, arr2):
	"""The Spearman correlation between two tensors across the last axis.

	Computes the Spearman correlation in the last dimension of `arr1` and `arr2`.
	`arr1` and `arr2` must be the same shape. For example, if they are both
	A x B x L arrays, then the correlation of corresponding L-arrays will be
	computed and returned in an A x B array.

	A dense ordering is used and ties are broken based on position in the
	tensor.


	Parameters
	----------
	arr1: torch.Tensor
		One of the tensor to correlate. This can be any number of dimensions but the MSE is
		calculated across the last dimension.

	arr2: torch.Tensor
		The other tensor to correlation.


	Returns
	-------
	correlation: torch.Tensor
		The correlation for each element.
	"""

	ranks1 = arr1.argsort().argsort().type(torch.float32)
	ranks2 = arr2.argsort().argsort().type(torch.float32)
	return pearson_corr(ranks1, ranks2)


def mean_squared_error(arr1, arr2):
	"""The mean squared error between two tensors averaged along the last axis.

	Computes the element-wise squared error between two tensors and averages
	these across the last dimension. `arr1` and `arr2` must be the same shape. 
	For example, if they are both A x B x L arrays, then the MSE of 
	corresponding L-arrays will be computed and returned in an A x B array.


	Parameters
	----------
	arr1: torch.Tensor
		A tensor of values.

	arr2: torch.Tensor
		Another tensor of values with the same shape as arr1.


	Returns
	-------
	mse: torch.Tensor
		The L2 distance between two tensors.
	"""

	return torch.mean(torch.square(arr1 - arr2), axis=-1)


def calculate_performance_measures(logps, true_counts, pred_log_counts,
	kernel_sigma=7, kernel_width=81, smooth_true=False, 
	smooth_predictions=False, measures=None):
	"""Calculates a set of performance measures given true and observed data.

	This function will take in observed readouts, predicted profiles, and
	predicted counts, and calculate a series of specified performance measures
	on them. Each performance measure could be calculated individually using
	its function, but this function provides a wrapper around running any
	number of them. The measures one can choose are:

		Profile Performance Measures
		- "profile_mnll": the multinomial log-likelihood of the observed profile given
			the predicted logits
		- "profile_jsd": the Jensen-Shannon divergence between the observed profile and
			the predicted probabilities
		- "profile_pearson": the Pearson correlation between the observed profile and
			the predicted probabilities
		- "profile_spearman": the Spearman correlation between the observed profiles
			and the predicted probabilities

		Count Performance Measures
		- "count_pearson": the Pearson correlation between the observed log counts and
			the predicted log counts
		- "count_spearman": the Spearman correlation between the observed log counts
			and the predicted log counts
		- "count_mse": the mean-squared error between the observed log counts and the
			predicted log counts.

	Optionally, one can choose to smooth the *observed* data before calculating
	the profile correlations and JSD. It is important to note that this smoothing
	is not being done on the predictions, but on the observed bp-resolution
	counts, with the reasoning being that these counts are sparse due to their
	bp resolution nature. The smoothing is done according to a Gaussian with a
	sigma and kernel width as specified.


	Parameters
	----------
	logps: torch.Tensor, shape=(n, n_strands, length)
		The predicted logits or log probabilities for each basepir for each strand. 
		If the predictions are unstranded, this dimension must be 1.

	true_counts: torch.Tensor, shape=(n, n_strands, length)
		The integer counts of the number of reads per basepair in the observed data.
	
	pred_log_counts: torch.Tensor, shape=(n, n_outputs)
		The predicted log counts for each example.

	kernel_sigma: int, optional
		If smoothing the observed profile, the sigma to use in the Gaussian
		smoothing. Default is 7.

	kernel_width: int, optional
		If smoothing the observed profile, the kernel width to use in the Gaussian
		smoothing. Default is 81.

	smooth_true: bool, optional
		Whether to smooth the observed data using a Gassian kernel. Default is False.

	smooth_predictions: bool, optional
		Whether to smooth the predicted values using a Gaussian kernel. Default is
		False.

	measures: None or list, optional
		If a list of strings, each string should correspond to a performance measure
		to calculate. If None, calculate all performance measures.


	Returns
	-------
	measures_: dict of torch.Tensors
		A dictionary where the keys are the names of performance measures and the
		values are tensors containing the values. Each profile performance measure
		will have the shape (n, 1) and each count performance measure will have
		the shape (1,).
	"""

	measures_ = {}
	logps = torch.nn.functional.log_softmax(logps, dim=-1)

	# Profile Performance Meausres
	if measures is None or 'profile_mnll' in measures: 
		measures_['profile_mnll'] = batched_smoothed_function(logps=logps, 
			true_counts=true_counts, f=MNLLLoss, 
			smooth_predictions=smooth_predictions, smooth_true=False, 
			kernel_sigma=kernel_sigma, kernel_width=kernel_width)

	if measures is None or 'profile_jsd' in measures: 
		measures_['profile_jsd'] = batched_smoothed_function(logps=logps, 
			true_counts=true_counts, f=jensen_shannon_distance, 
			smooth_predictions=smooth_predictions, smooth_true=smooth_true, 
			kernel_sigma=kernel_sigma, kernel_width=kernel_width)

	if measures is None or 'profile_pearson' in measures:
		measures_['profile_pearson'] = batched_smoothed_function(logps=logps, 
			true_counts=true_counts, f=pearson_corr, 
			smooth_predictions=smooth_predictions, smooth_true=smooth_true, 
			exponentiate_logps=True, kernel_sigma=kernel_sigma, 
			kernel_width=kernel_width)

	if measures is None or 'profile_spearman' in measures:
		measures_['profile_spearman'] = batched_smoothed_function(logps=logps, 
			true_counts=true_counts, f=spearman_corr, 
			smooth_predictions=smooth_predictions, smooth_true=smooth_true, 
			exponentiate_logps=True, kernel_sigma=kernel_sigma, 
			kernel_width=kernel_width)


	# Count Performance Measures
	true_counts = true_counts.sum(dim=(-1, -2)).unsqueeze(-1)
	true_log_counts = torch.log(true_counts + 1)

	if measures is None or 'count_pearson' in measures:
		measures_['count_pearson'] = pearson_corr(pred_log_counts.T, 
			true_log_counts.T)

	if measures is None or 'count_spearman' in measures:
		measures_['count_spearman'] = spearman_corr(pred_log_counts.T, 
			true_log_counts.T)

	if measures is None or 'count_mse' in measures:
		measures_['count_mse'] = mean_squared_error(pred_log_counts.T, 
			true_log_counts.T)

	return measures_



# bpnet-lite
# Author: Jacob Schreiber
# Code adapted from Avanti Shrikumar and Ziga Avsec and Alex Tseng and many others

from .bpnet import BPNet
from .chrombpnet import ChromBPNet

__version__ = '0.9.4'
